#include <ACS_ZDoom.h>
#include <stdlib.h>
#include <ctype.h>
#include <stdbool.h>
#include "utility.h"

char *astrtoc(__str s)
{
	static char *buf;
	static size_t alloc;
	size_t len = ACS_StrLen(s) + 1;
	if(alloc < len)
		buf = realloc(buf, len);
	for(size_t i = 0; i != len; ++i)
		buf[i] = s[i];
	return buf;
}

__str astrupper(__str s)
{
	int len = ACS_StrLen(s);
	ACS_BeginStrParam();
	for(int i = 0; i < len; i++)
	{
		char c = ACS_GetChar(s, i);
		__nprintf("%c", toupper(c));
	}
	return ACS_EndStrParam();
}

__str astrlower(__str s)
{
	int len = ACS_StrLen(s);
	ACS_BeginStrParam();
	for(int i = 0; i < len; i++)
	{
		char c = ACS_GetChar(s, i);
		__nprintf("%c", tolower(c));
	}
	return ACS_EndStrParam();
}

int astrtoi(__str s)
{
	return strtol_str(s, NULL, 10);
}

void FreezePlayer()
{
	ACS_SetActorVelocity(0, 0, 0, ACS_GetActorVelZ(0), false, true);
	ACS_SetPlayerProperty(0, 1, PROP_TOTALLYFROZEN);
}

void UnFreezePlayer()
{
	ACS_SetPlayerProperty(0, 0, PROP_TOTALLYFROZEN);
}

double GetAspectRatio()
{
	int aspect        = ACS_GetCVar(s"vid_aspect");
	double width      = ACS_GetCVarFixed(s"vid_defwidth");
	double height     = ACS_GetCVarFixed(s"vid_defheight");
	bool nowidescreen = ACS_GetCVar(s"vid_nowidescreen");
	bool tft          = ACS_GetCVar(s"vid_tft");
	
	if(nowidescreen)
	{
		if (!tft)
			return ASPECT_4_3;
		if (height * ASPECT_5_4 == width)
			return ASPECT_5_4;
		else
			return ASPECT_4_3;
	}
	
	switch(aspect)
	{
		case 1: return ASPECT_16_9;
		case 2: return ASPECT_16_10;
		case 3: return ASPECT_4_3;
		case 4: return ASPECT_5_4;
		case 5: return ASPECT_17_10;
	}
	
	return width / height;
}

double GetAspectScale()
{
	// Return ASPECT_4_3 for crop
	// Return GetAspectRatio() / ASPECT_4_3 for scale
	return 1 / (GetAspectRatio() / ASPECT_4_3);
}

bool KeyPressed(int k)
{
	int input     = ACS_GetPlayerInput(-1, INPUT_BUTTONS);
	int old_input = ACS_GetPlayerInput(-1, INPUT_OLDBUTTONS);
    return ((input & k) && !(old_input & k));
}

bool KeyReleased(int k)
{
	int input     = ACS_GetPlayerInput(-1, INPUT_BUTTONS);
	int old_input = ACS_GetPlayerInput(-1, INPUT_OLDBUTTONS);
    return (!(input & k) && (old_input & k));
}
