//
// VNS Reader
//

#define ACS_SHORT_NAMES

#include <ACS_ZDoom.h>
#include <stdlib.h>
#include <stdbool.h>
#include "defs.h"
#include "utility.h"
#include "vector.h"
#include "vns.h"

#define HID_BASE      100
#define HID_PORTRAIT  (HID_BASE - 1)
#define HID_PORTRAITF (HID_BASE - 2)
#define HID_CG        (HID_BASE - 3)
#define HID_CGF       (HID_BASE - 4)
#define HID_INTERFACE (HID_BASE - 5)
#define HID_NAME      (HID_BASE - 6)
#define HID_SPEAKING  (HID_BASE - 7)
#define HID_DIALOGUE  (HID_BASE - 8)

static vns_t *vns = NULL;
[[call("SScriptS")]] static opfunc_t opcodes[] =
{
	[OP_NOP]          = &opNop,
	[OP_NAME]         = &opName,
	[OP_PORTRAIT]     = &opPortrait,
	[OP_FONT]         = &opFont,
	[OP_TEXT]         = &opText,
	[OP_CHOICE_START] = &opChoiceStart,
	[OP_CHOICE_ADD]   = &opChoiceAdd,
	[OP_CHOICE_END]   = &opChoiceEnd,
	[OP_END]          = &opEnd,
	[OP_SCRIPT]       = &opScript,
	[OP_WAITTEXT]     = &opWaitText,
	[OP_WAIT]         = &opWait,
};

[[extern("ACS"), call("ScriptS")]]
void VNS_Execute(int s)
{
	// stuff needs to be loaded first
	while (vns == NULL || !vns->loaded)
		Delay(1);
	
	int pn = PlayerNumber();
	int tid = ActivatorTID();
	if (pn == -1)
	{
		if (!tid)
			Thing_ChangeTID(0, UniqueTID(MIN_TID, 0));
		tid = ActivatorTID();
		SetActivatorToTarget(0);
		pn = PlayerNumber();
		s  = GetUserArray(tid, s"user_vns_script", pn);
	}
	
	if (s >= vns->num_scripts || s < 0)
	{
		Log("VNS_Execute: Unknown script %d", s);
		SetUserVariable(tid, s"user_vns_active", false);
		return;
	}
	
	script_t *scr = &vns->scripts[s];
	for (int i = 0; i < scr->num_actions; i++)
	{
		if (scr->actions[i].opcode < sizeof(opcodes))
		{
			int r = opcodes[scr->actions[i].opcode](scr, pn, tid, scr->actions[i].args, scr->actions[i].argv);
			if (r > 0)
				break;
		}
	}
	
	SetUserVariable(tid, s"user_vns_active", false);
}

[[call("ScriptS"), script("Enter")]]
void VNS_Handler(void)
{
	if (GameType() == GAME_TITLE_MAP)
		return;
	
	__str lastPortrait, lastWeapon;
	int width, height, tics;
	int    pn        = PlayerNumber();
	double textSpeed = 0.05;
	while (true)
	{
		if (vns->frames[pn].count > 0)
		{
			// Freeze player and holster weapon
			FreezePlayer();
			lastWeapon = GetWeapon();
			GiveInventory(s"Holster", 1);
			SetWeapon(s"Holster");
			
			vnframe_t *f = vns->frames[pn].at(&vns->frames[pn], 0);
			lastPortrait = f->portrait;
			
			// Fade everything in
			// Interface
			SetHudSize(320, 240, true);
			SetFont(s"VN_INTFC");
			HudMessage(HUDMSG_FADEINOUT, HID_INTERFACE, CR_UNTRANSLATED, .5, .5, 0, 1, 0, 0, "A");
			
			// Name & Speaking
			SetHudSize(480, 280, true);
			SetFont(s"SMALLFONT");
			HudMessage(HUDMSG_FADEINOUT, HID_NAME, CR_UNTRANSLATED, 7.5, 194.2, 0, 1, 0, 0, "%S", f->name);
			HudMessage(HUDMSG_FADEINOUT, HID_SPEAKING, CR_UNTRANSLATED, 345.1, 193.2, 0, 1, 0, 0, "S P E A K I N G");
			
			// Portrait
			width      = 1600 * GetAspectScale();
			height     = 900;
			SetHudSize(width, height, true);
			SetFont(f->portrait);
			HudMessage(HUDMSG_FADEINOUT, HID_PORTRAIT, CR_UNTRANSLATED, (int)(width / 2), (int)(height / 2), 0, 1, 0, 0, !StrLen(f->portrait) ? "" : "A");
			
			for (int i = 0; i < 35; i++)
			{
				Delay(1);
				if (KeyPressed(BT_USE))
					break;
			}
			
			while (vns->frames[pn].count > 0)
			{
				FreezePlayer();
				f = vns->frames[pn].at(&vns->frames[pn], 0);
				
				switch(f->type)
				{
					default:
					case VNS_DIALOGUE:
					{
						tics = 35 * textSpeed * StrLen(f->text);
						
						// portrait transition
						if (StrICmp(lastPortrait, f->portrait))
						{
							width      = 1600 * GetAspectScale();
							height     = 900;
							SetHudSize(width, height, true);
							SetFont(f->portrait);
							HudMessage(HUDMSG_FADEINOUT, HID_PORTRAITF, CR_UNTRANSLATED, (int)(width / 2), (int)(height / 2), 0, 1, 0, 0, !StrLen(f->portrait) ? "" : "A");
							
							for (int i = 0; i < 35; i++)
							{
								Delay(1);
								if (KeyPressed(BT_USE))
									break;
							}
							
							ClearMessage(HID_PORTRAITF);
						}
						
						// Interface
						SetHudSize(320, 240, true);
						SetFont(s"VN_INTFC");
						HudMessage(HUDMSG_PLAIN, HID_INTERFACE, CR_UNTRANSLATED, .5, .5, 0, 0, 0, 0, "A");
						
						// Name & Speaking
						SetHudSize(480, 280, true);
						SetFont(s"SMALLFONT");
						HudMessage(HUDMSG_PLAIN, HID_NAME, CR_UNTRANSLATED, 7.5, 194.2, 0, 0, 0, 0, "%S", f->name);
						HudMessage(HUDMSG_PLAIN, HID_SPEAKING, CR_UNTRANSLATED, 345.1, 193.2, 0, 0, 0, 0, "S P E A K I N G");
						
						// Dialogue
						SetFont(f->font);
						HudMessage(HUDMSG_TYPEON, HID_DIALOGUE, CR_WHITE, 70.1, 205.1, 0, textSpeed, 0, 0, "%S", f->text);
						
						for (int i = 0; i < tics; i++)
						{
							// constantly draw portrait so that it scales when the resolution changes
							width      = 1600 * GetAspectScale();
							height     = 900;
							SetHudSize(width, height, true);
							SetFont(f->portrait);
							HudMessage(HUDMSG_PLAIN, HID_PORTRAIT, CR_UNTRANSLATED, (int)(width / 2), (int)(height / 2), 0, 0, 0, 0, !StrLen(f->portrait) ? "" : "A");
							
							Delay(1);
							if (KeyPressed(BT_USE))
								break;
						}
						
						SetHudSize(480, 280, true);
						SetFont(f->font);
						HudMessage(HUDMSG_PLAIN, HID_DIALOGUE, CR_WHITE, 70.1, 205.1, 0, 0, 0, 0, "%S", f->text);
						
						while (true)
						{
							// constantly draw portrait so that it scales when the resolution changes
							width      = 1600 * GetAspectScale();
							height     = 900;
							SetHudSize(width, height, true);
							SetFont(f->portrait);
							HudMessage(HUDMSG_PLAIN, HID_PORTRAIT, CR_UNTRANSLATED, (int)(width / 2), (int)(height / 2), 0, 0, 0, 0, !StrLen(f->portrait) ? "" : "A");
							
							Delay(1);
							if (KeyPressed(BT_USE))
								break;
						}
					}
					
					case VNS_CHOICE:
						break;
				}
				
				lastPortrait = f->portrait;
				vns->frames[pn].remove(&vns->frames[pn], 0);
			}
			
			// Fade everything out
			// Interface
			SetHudSize(320, 240, true);
			SetFont(s"VN_INTFC");
			HudMessage(HUDMSG_FADEINOUT, HID_INTERFACE, CR_UNTRANSLATED, .5, .5, 0, 0, 1, 0, "A");
			
			// Name & Speaking
			SetHudSize(480, 280, true);
			SetFont(s"SMALLFONT");
			HudMessage(HUDMSG_FADEINOUT, HID_NAME, CR_UNTRANSLATED, 7.5, 194.2, 0, 0, 1, 0, "%S", f->name);
			HudMessage(HUDMSG_FADEINOUT, HID_SPEAKING, CR_UNTRANSLATED, 345.1, 193.2, 0, 0, 1, 0, "S P E A K I N G");
			
			// Text
			SetFont(f->font);
			HudMessage(HUDMSG_FADEINOUT, HID_DIALOGUE, CR_WHITE, 70.1, 205.1, 0, 0, 1, 0, "%S", f->text);
			
			// Portrait
			width      = 1600 * GetAspectScale();
			height     = 900;
			SetHudSize(width, height, true);
			SetFont(f->portrait);
			HudMessage(HUDMSG_FADEINOUT, HID_PORTRAIT, CR_UNTRANSLATED, (int)(width / 2), (int)(height / 2), 0, 0, 1, 0, !StrLen(f->portrait) ? "" : "A");
			
			for (int i = 0; i < 35; i++)
			{
				Delay(1);
				if (KeyPressed(BT_USE))
					break;
			}
			
			ClearMessage(HID_INTERFACE);
			ClearMessage(HID_NAME);
			ClearMessage(HID_SPEAKING);
			ClearMessage(HID_DIALOGUE);
			ClearMessage(HID_PORTRAIT);
			
			UnFreezePlayer();
			SetWeapon(lastWeapon);
			TakeInventory(s"Holster", 1);
		}
		
		Delay(1);
	}
}

[[call("ScriptS"), script("Open")]]
void VNS_Initialize(void)
{
	if (GameType() == GAME_TITLE_MAP)
		return;
	
	vns = (vns_t *)malloc(sizeof(vns_t));
	vns->loaded   = false;
	vns->name     = (__str *)calloc(MAX_PLAYERS, sizeof(__str));
	vns->portrait = (__str *)calloc(MAX_PLAYERS, sizeof(__str));
	vns->font     = (__str *)calloc(MAX_PLAYERS, sizeof(__str));
	vns->frames   = (vector_t *)calloc(MAX_PLAYERS, sizeof(vector_t));
	
	for (int i = 0; i < MAX_PLAYERS; i++)
	{
		vns->name[i]     = s"";
		vns->portrait[i] = s"";
		vns->font[i]     = s"SMALLFONT";
		vns->frames[i]   = *new_vector(sizeof(vnframe_t));
	}
	
	vns->num_scripts = astrtoi(StrParamL("VNS_NUM_SCRIPTS"));
	vns->scripts     = (script_t *)calloc(vns->num_scripts, sizeof(script_t));
	for (int s = 0; s < vns->num_scripts; s++)
	{
		__str sn = StrParamL("VNS_SCRIPT_%d", s);
		
		script_t *scr = (script_t *)malloc(sizeof(script_t));
		scr->name     = sn;
		
		scr->num_strings = astrtoi(StrParamL("VNS_%S_STRING_TABLE_L", sn));
		scr->num_actions = astrtoi(StrParamL("VNS_%S_ACTION_TABLE_L", sn));
		scr->strings = (__str *)calloc(scr->num_strings, sizeof(__str));
		scr->actions = (action_t *)calloc(scr->num_actions, sizeof(action_t));
		
		for (int i = 0; i < scr->num_strings; i++)
			scr->strings[i] = StrParamL("VNS_%S_STRING_ENTRY_%d", sn, i);
		for (int i = 0; i < scr->num_actions; i++)
		{
			scr->actions[i].opcode = astrtoi(StrParamL("VNS_%S_ACTION_%d_OPCODE", sn, i));
			scr->actions[i].args   = astrtoi(StrParamL("VNS_%S_ACTION_%d_ARGS", sn, i));
			scr->actions[i].argv   = (int *)calloc(scr->actions[i].args, sizeof(int));
			for (int a = 0; a < scr->actions[i].args; a++)
				scr->actions[i].argv[a] = astrtoi(StrParamL("VNS_%S_ACTION_%d_ARG_%d", sn, i, a));
		}
		
		vns->scripts[s] = *scr;
		
		// we don't wanna hit the action limit
		Delay(1);
	}
	
	vns->loaded = true;
}

[[call("SScriptS")]] int opNop(script_t *s, int pn, int tid, int argc, int *argv)
{
	return 0;
}

[[call("SScriptS")]] int opName(script_t *s, int pn, int tid, int argc, int *argv)
{
	vns->name[PlayerNumber()] = s->strings[argv[0]];
	return 0;
}

[[call("SScriptS")]] int opPortrait(script_t *s, int pn, int tid, int argc, int *argv)
{
	vns->portrait[pn] = StrParamL("%S", s->strings[argv[0]]);
	return 0;
}

[[call("SScriptS")]] int opFont(script_t *s, int pn, int tid, int argc, int *argv)
{
	vns->font[pn] = s->strings[argv[0]];
	return 0;
}

[[call("SScriptS")]] int opText(script_t *s, int pn, int tid, int argc, int *argv)
{
	if (argc < 2 || pn == -1)
		return 1;
	
	vnframe_t *f = (vnframe_t *)malloc(sizeof(vnframe_t));
	f->type      = VNS_DIALOGUE;
	f->text      = s->strings[argv[0]];
	f->words     = argv[1];
	f->portrait  = vns->portrait[pn];
	f->name      = vns->name[pn];
	f->font      = vns->font[pn];
	vns->frames[pn].push_back(&vns->frames[pn], f);
	
	return 0;
}

[[call("SScriptS")]] int opChoiceStart(script_t *s, int pn, int tid, int argc, int *argv)
{
	Log("choice_start: %S", s->strings[argv[0]]);
	return 0;
}

[[call("SScriptS")]] int opChoiceAdd(script_t *s, int pn, int tid, int argc, int *argv)
{
	Log("choice_add: %S", s->strings[argv[0]]);
	return 0;
}

[[call("SScriptS")]] int opChoiceEnd(script_t *s, int pn, int tid, int argc, int *argv)
{
	Log("choice_end");
	return 0;
}

[[call("SScriptS")]] int opEnd(script_t *s, int pn, int tid, int argc, int *argv)
{
	return 1;
}

[[call("SScriptS")]] int opScript(script_t *s, int pn, int tid, int argc, int *argv)
{
	for (int i = 0; i < vns->num_scripts; i++)
	{
		if (!StrICmp(vns->scripts[i].name, s->strings[argv[0]]))
		{
			SetUserArray(tid, s"user_vns_script", pn, i);
			break;
		}
	}
	return 0;
}

[[call("SScriptS")]] int opWaitText(script_t *s, int pn, int tid, int argc, int *argv)
{
	while (vns->frames[pn].count > 0)
		Delay(1);
	return 0;
}

[[call("SScriptS")]] int opWait(script_t *s, int pn, int tid, int argc, int *argv)
{
	Delay(argv[0]);
	return 0;
}
